/*
 * cycfx2prog.cc - Cypress FX2(LP) programmer. 
 * 
 * Copyright (c) 2006--2009 by Wolfgang Wieser ] wwieser (a) gmx <*> de [ 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include "cycfx2dev.hh"

// The version is set in Makefile.
#define  CYCFX2PROG_VERSION 0.47

#ifndef CYCFX2PROG_VERSION
# error CYCFX2PROG_VERSION undefined
#endif

// Global FX2 device we're connected to. 
CypressFX2Device cycfx2;

extern "C" void resetfx2_(void);
extern "C" void runfx2_(void);
extern "C" void prgfx2_(char *);
extern "C" int get_fx2buff_(int, unsigned char *);
extern "C" int write_fx2buff_(char *);
extern "C" int initfx2_(void);
extern "C" int cmd_proc_(int,char *);

static void *CheckMalloc(void *ptr)
{
	if(!ptr)
	{
		fprintf(stderr,"malloc failed\n");
		exit(1);
	}
	return(ptr);
}


// FIXME: What to do with these: 
// HMMM.... There is usb_device::num_children and **usb_device::children. 

void USBDumpBusses(FILE *out)
{
	for(usb_bus *b=usb_busses; b; b=b->next)
	{
		for(struct usb_device *d=b->devices; d; d=d->next)
		{
			bool is_fx2_dev = (d->descriptor.idVendor==0x4b4 && 
				d->descriptor.idProduct==0x8613);
			fprintf(out,"Bus %s Device %s: ID %04x:%04x%s\n",
				b->dirname,d->filename,
				d->descriptor.idVendor,d->descriptor.idProduct,
				is_fx2_dev ? " (unconfigured FX2)" : "");
		}
	}
}


void resetfx2_(void){
cycfx2.FX2Reset(0); 
}

void prgfx2_(char * p){
cmd_proc_(1,p);
}
void runfx2_(void){
cycfx2.FX2Reset(1);
}

int get_fx2buff_(int len, unsigned char *buf){
  int ep=0x86;
  char type='B';
  int rv=cycfx2.BlockRead(ep,buf,len,type);
  return rv;
}

int write_fx2buff_(char *buf){
cmd_proc_(1,buf);
return (0);
}

int initfx2_(void)
{

	const char *arg_bus_dev=NULL;

	int arg_vend=0x4b4,arg_prod=0x8613,arg_nth=0,arg_id_specified=0;

	// Initialize the USB library...
	usb_init();
	
	int rv=usb_find_busses();
	if(rv<0)
	{  fprintf(stderr,"usb_find_busses failed (rv=%d)\n",rv);  return(1);  }
	rv=usb_find_devices();
	if(rv<0)
	{  fprintf(stderr,"usb_find_devices failed (rv=%d)\n",rv);  return(1);  }
	
	// Look for device. 
	struct usb_device *usbdev=NULL;
	if(arg_bus_dev)
	{
		char bus[16];
		char dev[16];
		do {
			if(strlen(arg_bus_dev)>=16)  break;  // Prevents buffer overflows. 
			const char *p=strchr(arg_bus_dev,'.');
			if(!p)  break;
			strncpy(bus,arg_bus_dev,p-arg_bus_dev);
			bus[p-arg_bus_dev]='\0';
			strcpy(dev,p+1);
			usbdev=USBFindDevice(bus,dev);
		} while(0);
		if(!usbdev)
		{  fprintf(stderr,"Illegal/nonexistant device %s.\n",arg_bus_dev);
			return(1);  }
	}
	else
	{
		usbdev=USBFindDevice(arg_vend,arg_prod,arg_nth);
		if(!usbdev)
		{
			if(arg_id_specified)
			{  fprintf(stderr,"Device with vendorID=0x%04x, productID="
				"0x%04x, nth=%d not attached.\n",arg_vend,arg_prod,arg_nth);  }
			else
			{  fprintf(stderr,"No unconfigured Cypress FX2 attached.\n");  }
			return(1);
		}
	}
	
	fprintf(stderr,"Using ID %04x:%04x on %s.%s.\n",
		usbdev->descriptor.idVendor,usbdev->descriptor.idProduct,
		usbdev->bus->dirname,usbdev->filename);
	
	if(cycfx2.open(usbdev))
	{  return(1);  }

  return (0);
}

int cmd_proc_(int argc,char *cmdl)
{
	int errors=0;
// Execute all the user commands. 
	
	{
		
		// Copy command and args so that we can manipulate it. 
		char *cmd=(char*)CheckMalloc(malloc(strlen(cmdl)+1));

		strcpy(cmd,cmdl);
		
		// Cut into command and arguments: 
		const int MAXARGS=16;
		char *a[MAXARGS];
		for(int j=0; j<MAXARGS; j++)  a[j]=NULL;
		char *first_arg=strchr(cmd,':');
		int nargs=0;
		if(first_arg)
		{
			do {
				*(first_arg++)='\0';
				if(nargs>=MAXARGS)
				{
					fprintf(stderr,"Too many arguments for command \"%s\" "
						"(further args ignored)\n",cmd);
					++errors;
					break;
				}
				a[nargs++]=first_arg;
				first_arg=strchr(first_arg,',');
			} while(first_arg);
		}
		
#if 0
		// Debug: 
		printf("Command: <%s>",cmd);
		for(int j=0; j<nargs; j++)
		{  printf(" <%s>",a[j]);  }
		printf("\n");
#endif
		
		if(!strcmp(cmd,"reset"))
		{
			fprintf(stderr,"Putting 8051 into reset.\n");
			errors+=cycfx2.FX2Reset(/*running=*/0);
		}
		else if(!strcmp(cmd,"run"))
		{
			fprintf(stderr,"Putting 8051 out of reset.\n");
			errors+=cycfx2.FX2Reset(/*running=*/1);
		}
		else if(!strcmp(cmd,"prg"))
		{
			// NOTE: We put the 8051 into reset prior to downloading but 
			//       we won't put it out of reset afterwards. 
			fprintf(stderr,"Putting 8051 into reset.\n");
			errors+=cycfx2.FX2Reset(/*running=*/0);

			const char *file=a[0];
			printf(" PROG = %s\n",file);
			if(!file)
			{  fprintf(stderr,"Command \"dl\" requires file to download.\n");
				++errors;  }
			else
			{
				fprintf(stderr,"Programming 8051 using \"%s\".\n",file);
				errors+=cycfx2.ProgramIHexFile(file);
			}
		}
		else if(!strcmp(cmd,"delay"))
		{
			long delay=-1;
			if(a[0] && *a[0])
			{  delay=strtol(a[0],NULL,0);  }
			if(delay<0)  delay=250;
			fprintf(stderr,"Delay: %ld msec\n",delay);
			usleep(delay*1000);
		}
		else if(!strcmp(cmd,"dram"))
		{
			int adr=0;
			int len=1;
			if(a[0] && *a[0])
			{  adr=strtol(a[0],NULL,0);  }
			if(adr<0)  adr=0;
			if(a[1] && *a[1])
			{  len=strtol(a[1],NULL,0);  }
			if(len<1)  len=1;
			if(len>1024*1024)  len=1024*1024;
			
			fprintf(stderr,"Dumping %u bytes of RAM at 0x%x:\n",len,adr);
			unsigned char *buf=(unsigned char*)CheckMalloc(malloc(len));
			memset(buf,0,len);
			errors+=cycfx2.ReadRAM(adr,buf,len);
			if(buf)  free(buf);
		}
		else if(!strcmp(cmd,"set"))
		{
			int adr=-1;
			int val=-1;
			if(a[0] && *a[0])  {  adr=strtol(a[0],NULL,0);  }
			if(a[1] && *a[1])  {  val=strtol(a[1],NULL,0);  }
			if(adr<0 || val<0 || val>=256)
			{  fprintf(stderr,"Command set: Illegal/missing address "
				"and/or value.\n");  ++errors;  }
			else
			{
				fprintf(stderr,"Setting value at 0x%x to 0x%x\n",adr,val);
				unsigned char cval=val;
				errors+=cycfx2.WriteRAM(adr,&cval,1);
			}
		}
		else if(!strcmp(cmd,"dbuf"))
		{
			int ep=6;
			int len=512;
			int num=1;
			char type='b';
			// If len<0, allow short reads. 
			if(len<0)
			{  type='B';  len=-len;  }
			if(ep<0 || ep>=127 || len<=0 || len>32*1024*1024 || num<1)
			{  fprintf(stderr,"Command dbulk: Illegal/missing "
				"endpoint/length/number.\n");  ++errors;  }
			else
			{
				// IN endpoints have the bit 7 set. 
				ep|=0x80;
				
				unsigned char *buf=(unsigned char*)CheckMalloc(malloc(len));
				memset(buf,0,len);
				
				for(int i=0; i<num; i++)
				{
					fprintf(stderr,"Reading %s%d bytes from EP adr 0x%02x\n",
						type=='B' ? "<=" : "",len,ep);
					int rv=cycfx2.BlockRead(ep,buf,len,type);
					errors += rv<0 ? 1 : 0;
				}
				
				if(buf)  free(buf);
			}
		}

		else if(!strcmp(cmd,"dbulk"))
		{
			int ep=-1;
			int len=512;
			int num=1;
			char type='b';
			
			if(a[0] && *a[0])  {  ep=strtol(a[0],NULL,0);  }
			if(a[1] && *a[1])  {  len=strtol(a[1],NULL,0);  }
			if(a[2] && *a[2])  {  num=strtol(a[2],NULL,0);  }
			// If len<0, allow short reads. 
			if(len<0)
			{  type='B';  len=-len;  }
			if(ep<0 || ep>=127 || len<=0 || len>32*1024*1024 || num<1)
			{  fprintf(stderr,"Command dbulk: Illegal/missing "
				"endpoint/length/number.\n");  ++errors;  }
			else
			{
				// IN endpoints have the bit 7 set. 
				ep|=0x80;
				
				unsigned char *buf=(unsigned char*)CheckMalloc(malloc(len));
				memset(buf,0,len);
				
				for(int i=0; i<num; i++)
				{
					fprintf(stderr,"Reading %s%d bytes from EP adr 0x%02x\n",
						type=='B' ? "<=" : "",len,ep);
					int rv=cycfx2.BlockRead(ep,buf,len,type);
					errors += rv<0 ? 1 : 0;
				}
				
				if(buf)  free(buf);
			}
		}
		else if(!strcmp(cmd,"sbulk"))
		{
			int ep=-1;
			char type='b';
			const char *str=a[1];
			int len=str ? strlen(str) : 0;
			
			if(a[0] && *a[0])  {  ep=strtol(a[0],NULL,0);  }
			if(ep<0 || ep>=127)
			{  fprintf(stderr,"Command sbulk: Illegal/missing "
				"endpoint.\n");  ++errors;  }
			else
			{
				fprintf(stderr,"Sending %d bytes to EP adr 0x%02x\n",len,ep);
				int rv=cycfx2.BlockWrite(ep,(const unsigned char*)str,len,type);
				errors += rv<0 ? 1 : 0;
			}
		}
		else if(!strcmp(cmd,"fbulk"))
		{
			char type='b';
			const char *filename = a[1];
			int ep=-1;
			int chunk_size=64;
			
			do {
				if(a[0] && *a[0])  {  ep=strtol(a[0],NULL,0);  }
				if(ep<0 || ep>=127)
				{  fprintf(stderr,"Command fbulk: Illegal/missing "
					"endpoint.\n");  ++errors;  break;  }
				
				if(a[2] && *a[2])  {  chunk_size=strtol(a[2],NULL,0);  }
				if(chunk_size<1 || chunk_size>2048)
				{  fprintf(stderr,"Command fbulk: Illegal chunk size %d.\n",
					chunk_size);  ++errors;  break;  }
				
				FILE *stream = filename ? fopen(filename,"r") : stdin;
				if(!stream)
				{
					fprintf(stderr,"Failed to open \"%s\" for reading: %s\n",
						filename ? filename : "[stdin]",strerror(errno));
					++errors;  break;
				}
				
				int len;
				int tot_bytes=0;
				unsigned char buf[chunk_size];
				
				fprintf(stderr,"Sending \"%s\" in chunks of %d bytes to EP adr "
					"0x%02x\n",
					filename ? filename : "[stdin]",chunk_size,ep);
				while((len=fread(buf,1,chunk_size,stream))>0)
				{
			    	int rv=cycfx2.BlockWrite(ep,buf,len,type);
					if(rv<0)
					{  ++errors; break;  }
					tot_bytes += len;
				}
				if(filename) fclose(stream);
				fprintf(stderr,"Sent %d bytes to EP adr 0x%02x\n",tot_bytes,ep);
			} while(0);
		}
		else if(!strcmp(cmd,"bench_bulk"))
		{
			int ep=-1;
			int len=1024*1024;
			int cs=65536;
			
			if(a[0] && *a[0])  {  ep=strtol(a[0],NULL,0);  }
			if(a[1] && *a[1])  {  len=strtol(a[1],NULL,0);  }
			if(a[2] && *a[2])  {  cs=strtol(a[2],NULL,0);  }
			// If len<0, allow short reads. 
			if(ep<0 || ep>=127 || len<=0 || len>32*1024*1024 || 
				cs<1 || cs>32*1024*1024)
			{  fprintf(stderr,"Command bench_bulk: Illegal/missing "
				"endpoint/length/number.\n");  ++errors;  }
			else
			{
				// IN endpoints have the bit 7 set. 
				ep|=0x80;
				
				int rv=cycfx2.BenchBlockRead(ep,len,cs,'b');
				errors += rv ? 1 : 0;
			}
		}
		else if(!strcmp(cmd,"altif"))
		{
			int af=-1;
			if(a[0] && *a[0])  {  af=strtol(a[0],NULL,0);  }
			
			cycfx2.ForceAltInterface(af);
		}
		else if(!strcmp(cmd,"ctrl"))
		{
			int requesttype=0,request=0;
			int value=0,index=0;
			if(a[0] && *a[0]) requesttype=strtol(a[0],NULL,0);
			if(a[1] && *a[1]) request=strtol(a[1],NULL,0);
			if(a[2] && *a[2]) value=strtol(a[2],NULL,0);
			if(a[3] && *a[3]) index=strtol(a[3],NULL,0);
			fprintf(stderr,"Sending control message type 0x%02x, request "
				"0x%02x (value=%d,index=%d)\n",
				requesttype,request,value,index);
			errors+=cycfx2.CtrlMsg(requesttype,request,value,index);
		}
		else
		{
			fprintf(stderr,"Ignoring unknown command \"%s\".\n",cmd);
			++errors;
		}
		
		if(cmd)
		{  free(cmd);  cmd=NULL;  }
	}
	
	return errors;
}
